/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:55:06 $
#ifndef BITVEC_H
#define BITVEC_H

#include <iostream.h>
#include <assert.h>

// Implement a bit-vector. This class can be replaced with any standard library class that 
// provides the same functionality. Many of the normal bit-wise operators are available on
// BitVec's.
// BitVec's clients are StateSet and the item set's (ISImpl).

class BitVec {
public:
	// Constructors, destructors, operator=:

	// Default operator creates the emptyset, with little effort.
        // This assumes that delete [] 0 is okay (when it is deleted later).
	inline BitVec();

	// Normal copy constructor.
	BitVec( const BitVec& r );

	// Normal non-virtual destructor.
	inline ~BitVec();

        // Needs a special operator=.
	const BitVec& operator=( const BitVec& r );


	// Tests:

	// A special equality operator.
	int operator==( const BitVec& r ) const;

        // Inequality operator.
	inline int operator!=( const BitVec& r ) const;

	// Is there a bit set?
	int something_set() const;

	// Howmany bits are set?
	int bits_set() const;

	// Bit vector operators:

	// set a bit
	inline BitVec& set_bit( const int r );

	// unset a bit
	inline BitVec& unset_bit( const int r );

	BitVec& bitwise_or( const BitVec& r );

	BitVec& bitwise_and( const BitVec& r );

	BitVec& bitwise_unset( const BitVec& r );

	BitVec& bitwise_complement();

        // Bitwise containment.
	int contains( const BitVec& r ) const;

	// Is a bit set?
	inline int contains( const int r ) const;

	// Does *this have a bit in common with r?
	int something_common( const BitVec& r ) const;

	// Make *this the empty vector.
        void clear();

	// What is the set bit with the smallest index in *this?
	int smallest() const;


	// Some width related members:

	inline int width() const;

	void set_width( const int r );

	// Recycle this BitVec.
	inline void reincarnate();

	// Shift *this left, zero filling.
	BitVec& left_shift( const int r );

	// Append another bit-vector.
	BitVec& append( const BitVec& r );


	// Iterators:

	// Place the index of the first set bit in the iteration in reference r.
        // r == -1 if there is no first one.
	inline int iter_start( int& r ) const;

        // Is r the last set bit in an iteration sequence.
	inline int iter_end( int r ) const;

        // Place the next set bit, after r (in the iteration sequence), in reference r.
	int iter_next( int& r ) const;


	// Other special members:

	friend ostream& operator<<( ostream& os, const BitVec& r );

	// Structural invariant on the BitVec.
	inline int class_invariant() const;

private:
	// Some trivial helpers:

	// Compute the number of words required to hold st bits.	
	inline int words_required( const int st ) const;

        // Compute the word index of a particular bit r.
	inline int word_index( const int r ) const;

        // Compute the bit index (from the LSB) of a bit r.
	inline int bit_index( const int r ) const;


        // Actual representation:
	int bits_in_use;
	int words;
	unsigned int *data;

	// A class constant, used for bit-vector width.
	enum { bits_per_word = (sizeof( unsigned int ) * 8) };
};



// Now for the inline versions of the member functions:
inline BitVec::BitVec() :
		bits_in_use( 0 ),
		words( 0 ),
		data( 0 ) {
	assert( class_invariant() );
}

inline BitVec::~BitVec() {
	assert( class_invariant() );
	delete [] data;
}

inline void BitVec::reincarnate() {
	assert( class_invariant() );
        clear();
	bits_in_use = 0;
        assert( class_invariant() );
}

inline int BitVec::words_required( const int st ) const {
	// This may be a little more complicated than it has to be.
	return( (st / bits_per_word) + ((st % bits_per_word) != 0 ? 1 : 0) );
}

inline int BitVec::word_index( const int r ) const {
	return( r / bits_per_word );
}

inline int BitVec::bit_index( const int r ) const {
	return( r % bits_per_word );
}

inline int BitVec::operator!=( const BitVec& r ) const {
	return( !operator==( r ) );
}

inline int BitVec::width() const {
	return( bits_in_use );
}

inline BitVec& BitVec::set_bit( const int r ) {
	assert( (0 <= r) && (r < width()) );
        assert( class_invariant() );
        data[word_index( r )] |= (1U << bit_index( r ));
	return( *this );
}

inline BitVec& BitVec::unset_bit( const int r ) {
	// Unset the r'th bit.
	assert( (0 <= r) && (r < width()) );
        assert( class_invariant() );
	data[word_index( r )] &= ~(1U << bit_index( r ));
	return( *this );
}

inline int BitVec::contains( const int r ) const {
	// Check if the r'th bit is set.
	assert( r < width() );
        assert( class_invariant() );
	return( (data[word_index( r )] & (1U << bit_index( r ))) != 0U );
}

inline int BitVec::iter_start( int& r ) const {
	assert( class_invariant() );
	return( r = smallest() );
}

inline int BitVec::iter_end( int r ) const {
	return( r == -1 );
}

inline int BitVec::class_invariant() const {
	return( (width() <= words * bits_per_word)
		&& (width() >= 0)
                && (words >= 0)
		&& (words == 0 ? data == 0 : data != 0) );
}

#endif

