/* (c) Copyright 1994 by Bruce Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:55:17 $
#ifndef CHARRANGE_H
#define CHARRANGE_H

#include "extras.h"
#include <assert.h>
#include <iostream.h>

// Represent a set of characters (in the alphabet) in a subrange of char.
// Used by most classes as the basis of regular expressions and automata.

class CharRange {
public:
	// Some constructors:
	// The default is the empty range of characters.
	//	and this does not satisfy the class invariant.
	inline CharRange();

	// Copy constructor is standard.
	inline CharRange( const CharRange& r );
	inline CharRange( const char l, const char h );
	inline CharRange( const char a );

	// Default operator=, destructor are okay

        // Some normal member functions:
        // Is char a in the range?
	inline int contains( const char a ) const;

	// Access to the representation:
	inline char low() const;
	inline char high() const;


        // Standard comparison operators.
	inline int operator==( const CharRange& r ) const;
	inline int operator!=( const CharRange& r ) const;

	// Containment operators.
	inline int operator<=( const CharRange& r ) const;
	inline int operator>=( const CharRange& r ) const;

	// Is there something in common between *this and r?
	inline int not_disjoint( const CharRange& r ) const;

	// Do *this and r overlap, or are they adjacent?
	inline int overlap_or_adjacent( const CharRange& r ) const;

	// Some operators on CharRanges.
	// Merge two CharRanges if they're adjacent:
	inline CharRange& merge( const CharRange& r );

	// Make *this the intersection of *this and r.
	inline CharRange& intersection( const CharRange& r );

	// Return the left and right excess of *this with r (respectively):
	inline CharRange left_excess( const CharRange& r ) const;
	inline CharRange right_excess( const CharRange& r ) const;


	// Define an ordering on CharRange's, used mainly in RE::ordering().
	int ordering( const CharRange& r ) const;


	// The class structural invariant.
	inline int class_invariant() const;

        // Some extra stuff.
	friend ostream& operator<<( ostream& os, const CharRange r );

private:
	char lo, hi;
};

// class_invariant() is not satisfied.
inline CharRange::CharRange() :
		lo( 1 ),
		hi( 0 ) {}

inline CharRange::CharRange( const CharRange& r ) :
		lo( r.lo ),
		hi( r.hi ) {
	assert( class_invariant() );
}

inline CharRange::CharRange( const char l, const char h ) :
		lo( l ),
		hi( h ) {
	assert( class_invariant() );
}

inline CharRange::CharRange( const char a ) :
		lo( a ),
		hi( a ) {
	assert( class_invariant() );
}

inline int CharRange::contains( const char a ) const {
	assert( class_invariant() );
	return( (lo <= a) && (a <= hi) );
}

inline char CharRange::low() const {
	assert( class_invariant() );
	return( lo );
}

inline char CharRange::high() const {
	assert( class_invariant() );
	return( hi );
}

inline int CharRange::operator==( const CharRange& r ) const {
	assert( class_invariant() );
	assert( r.class_invariant() );
	return( (lo == r.lo) && (hi == r.hi) );
}

inline int CharRange::operator!=( const CharRange& r ) const {
	assert( class_invariant() );
	assert( r.class_invariant() );
	return( !operator==( r ) );
}

inline int CharRange::operator<=( const CharRange& r ) const {
	assert( class_invariant() );
	assert( r.class_invariant() );
	return( r.lo <= lo && hi <= r.hi );
}

inline int CharRange::operator>=( const CharRange& r ) const {
	assert( class_invariant() );
	assert( r.class_invariant() );
	return( lo <= r.lo && r.hi <= hi );
}

inline int CharRange::not_disjoint( const CharRange& r ) const {
	assert( class_invariant() );
	assert( r.class_invariant() );
	return( (lo <= r.hi) && (hi >= r.lo) );
}

inline int CharRange::overlap_or_adjacent( const CharRange& r ) const {
	assert( class_invariant() );
	assert( r.class_invariant() );
	return( not_disjoint( r )
		|| (r.lo == hi + 1)
		|| (r.hi == lo - 1) );
}

inline CharRange& CharRange::merge( const CharRange& r ) {
	assert( overlap_or_adjacent( r ) );
	lo = (char)min( lo, r.lo );
	hi = (char)max( hi, r.hi );
	return( *this );
}

inline CharRange& CharRange::intersection( const CharRange& r ) {
	lo = (char)max( lo, r.lo );
	hi = (char)min( hi, r.hi );
        return( *this );
}

inline CharRange CharRange::left_excess( const CharRange& r ) const {
	auto CharRange ret( min( lo, r.lo ), max( lo, r.lo ) - 1 );
	return( ret );
}

inline CharRange CharRange::right_excess( const CharRange& r ) const {
	auto CharRange ret( min( hi, r.hi ) + 1, max( hi, r.hi ) );
	return( ret );
}

inline int CharRange::class_invariant() const {
	// This could possibly be wrong if something goes weird with char being
	// signed or unsigned.
	return( lo <= hi );
}

#endif

