/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/03 12:18:49 $
#include "crset.h"


CRSet::CRSet( const CharRange a ) :
		howmany( expansion_size ),
		in_use( 1 ),
		data( new CharRange [expansion_size] ) {
	data[0] = a;
	assert( class_invariant() );
}

CRSet::CRSet( const CRSet& r ) :
		howmany( r.in_use + expansion_size ),
		in_use( r.in_use ),
		data( new CharRange [r.in_use + expansion_size] ) {
	assert( r.class_invariant() );
	auto int i;
	for( i = 0; i < in_use; i++ ) {
		data[i] = r.data[i];
	}
        assert( class_invariant() );
}

const CRSet& CRSet::operator=( const CRSet& r ) {
	assert( r.class_invariant() );
        // May need to allocate some more memory.
	if( r.in_use > howmany ) {
		auto CharRange *d( new CharRange [howmany = r.in_use + expansion_size] );
		delete [] data;
                data = d;
	}
	assert( r.in_use <= howmany );

	auto int i;
	for( i = 0; i < r.in_use; i++ ) {
		data[i] = r.data[i];
	}
	in_use = r.in_use;

	assert( class_invariant() );
        return( *this );
}

void CRSet::add_at( CharRange a, int i ) {
	assert( class_invariant() );
	assert( a.class_invariant() );

        // Go through data[] to see if a can be coalesced with data[i].
	for( ; i < in_use && !data[i].not_disjoint( a ); i++ );

	// Perhaps this is a coalescing candidate.
	// There are several possibilities, namely:
	//	a and data[i] are disjoint
	//	a <= data[i] (a is entirely contained in data[i]
	//	data[i] <= a
	//	a and data[i] are not disjoint (but no containment holds).
	//		In this case, we have three pieces: the intersection,
	//		and a slice each from a, data[i].

	if( i < in_use ) {
		assert( data[i].not_disjoint( a ) );
		if( data[i] == a ) {
			// Nothing to do.
		} else {
			assert( data[i].not_disjoint( a ) && (data[i] != a) );

			// a needs some splitting. We know that the excess
			// does not fall into data[0..i], we begin adding the excess
			// at data[i+1].
			// For efficiency (not implemented here), we could just
			// append any part of data[i] that doesn't fall within
			// a, since data[i] is pairwise disjoint with all other 
			// elements in array data[].

			auto CharRange old( data[i] );
			data[i].intersection( a );
			//	data[i] now contains old intersection a.

			// Now deal with the left and right excesses of old with a.
			// There are a few possibilities:
			if( a.low() != old.low() ) {
				// Then there is a left excess 
				add_at( a.left_excess( old ), i + 1 );
			}
			if( a.high() != old.high() ) {
				// Then there is a right excess
				add_at( a.right_excess( old ), i + 1 );
			}
		}
	} else {
		assert( i >= in_use );

        	// There may still be something left to put in data[].
        	append( a );
	}

	assert( class_invariant() );
	return;
}

CRSet& CRSet::combine( const CRSet& r ) {
	assert( r.class_invariant() );
	assert( class_invariant() );
	auto int i;
	for( i = 0; i < r.in_use; i++ ) {
        	add( r.data[i] );
        }
	assert( class_invariant() );
        return( *this );
}

// Does *this and r cover the same set of letters?
int CRSet::equivalent_cover( const CRSet& r ) const {
	assert( class_invariant() );
	auto int i;
	auto int result( 1 );
	for( i = 0; (i < in_use) && result; i++ ) {
		result = r.covered( data[i] );
	}
        return( result );
}

// Is r covered by this set?
int CRSet::covered( const CharRange r ) const {
	assert( class_invariant() );
	assert( r.class_invariant() );

	auto int i;

	for( i = 0; (i < in_use) && !r.not_disjoint( data[i] ); i++ );

	if( i < in_use ) {
		assert( r.not_disjoint( data[i] ) );
		// They have something in common.
		if( r <= data[i] ) {
			// r is completely covered by data[i]
			return( 1 );
		} else {
			// Now, figure out if the excesses are
			// also covered.
			// This could be made more efficient!
			auto int ret1( 1 );
			auto int ret2( 1 );
			if( r.low() != data[i].low() ) {
				// There actually is a left excess:
				ret1 = covered( r.left_excess( data[i] ) );
			}
			if( r.high() != data[i].high() ) {
				// There actually is a right excess:
				ret2 = covered( r.right_excess( data[i] ) );
			}
			return( ret1 && ret2 );
		}
	} else {
		assert( i >= in_use );
		// r is not covered by *this.
		return( 0 );
	}
}

// Make sure that data is at least s (adjusting howmany, but not in_use).
void CRSet::ensure_min_size( int s ) {
	assert( class_invariant() );
        // May need to allocate some new memory.
	if( s > howmany ) {
		auto CharRange *d( new CharRange [howmany = s + expansion_size] );
		auto int i;
		for( i = 0; i < in_use; i++ ) {
			d[i] = data[i];
		}
		delete [] data;
                data = d;
	}

	assert( class_invariant() ); 
}

ostream& operator<<( ostream& os, const CRSet& r ) {
	os << '{';
	auto int i;
	for( i = 0; i < r.in_use; i++ ) {
		os << ' ' << r.data[i] << ' ';
	}
        return( os << '}' );
}
