/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:55:39 $
#ifndef CRSET_H
#define CRSET_H

#include "charrang.h"
#include <assert.h>
#include <iostream.h>


// A data-structure to hold sets of CharRange's. The CharRange's are kept pair-
// wise disjoint.
// These are generally used in the construction of DFA's, and in DFA minimization.

class CRSet {
public:
	// Some basic constructors.
	// Create the emptyset:
	inline CRSet();

	// The singleton set:
	CRSet( const CharRange a );

	CRSet( const CRSet& r );

	// Also need a destructor and an operator=
	inline ~CRSet();

	const CRSet& operator=( const CRSet& r );


	// Some basic operations.

	// Add a CharRange to this set, splitting it as required.
	// It's not possible to add the empty CharRange to a CRSet!
	inline CRSet& add( CharRange a );

	// Append a CharRange onto the end of this set, knowing that a is disjoint with
        // any others in the set.
        inline CRSet& append( const CharRange a );

	// Combine two CRSet's, splitting CharRange's to maintain their pairwise
	// disjointness.
	CRSet& combine( const CRSet& s );

	// Does *this and r cover the same set of characters?
	int equivalent_cover( const CRSet& r ) const;

	// How many CharRanges are in *this?
	inline int size() const;

	// Some iterators:
	// Fetch the it'th CharRange in *this.
	inline const CharRange& iterator( const int it ) const;
	// Is there even an it'th CharRange?
	inline int iter_end( const int it ) const;


	// Some special members:
	inline int class_invariant() const;

	friend ostream& operator<<( ostream& os, const CRSet& r );

private:
	// A simple helper.
	void ensure_min_size( int s );

	// Add a CharRange to this set, splitting it as required.
	// Adding begins at i. This is used in CRSet::add().
	void add_at( CharRange a, int i );

	// Is r covered by this set?
	int covered( const CharRange r ) const;

	// A class constant:
        //	must be at least 1.
	enum { expansion_size = 5 };

	// Some implementation details:
	int howmany;
	int in_use;
	CharRange *data;
};

inline CRSet::CRSet() :
		howmany( 0 ),
		in_use( 0 ),
		data( 0 ) {
	assert( class_invariant() );
}

inline CRSet::~CRSet() {
	assert( class_invariant() );
	delete [] data;
}

// Insert a without checking to see if it overlaps with any of them already.
inline CRSet& CRSet::append( const CharRange a ) {
	assert( class_invariant() );
	assert( a.class_invariant() );
	ensure_min_size( in_use + 1 );
	data[in_use++] = a;
        assert( class_invariant() );
        return( *this );
}

// Add a CharRange to this set, splitting it as required.
// It's not possible to add the empty CharRange to a CRSet!
inline CRSet& CRSet::add( CharRange a ) {
	assert( class_invariant() );
        add_at( a, 0 );
	return( *this );
}

inline int CRSet::size() const {
	assert( class_invariant() );
	return( in_use );
}

inline const CharRange& CRSet::iterator( const int it ) const {
	assert( class_invariant() );
	assert( !iter_end( it ) );
	return( data[it] ); 
}

inline int CRSet::iter_end( const int it ) const {
	assert( class_invariant() );
	assert( 0 <= it );
	return( it >= in_use );
}

inline int CRSet::class_invariant() const {
	auto int result;
	result =  0 <= howmany
        	&& 0 <= in_use
		&& in_use <= howmany
		&& ( howmany == 0 ? data == 0 : data != 0 )
		&& 1 <= expansion_size;

        // Make sure that all of data[] are pairwise disjoint.
	auto int i;
	for( i = 0; (i < in_use) && result; i++ ) {
		auto int j;
		for( j = i + 1; (j < in_use) && result; j++ ) {
                	result = !data[i].not_disjoint( data[j] );
                }
	}
        return( result );
}

#endif

