/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:55:57 $
#include "re.h"
#include "sigma.h"


// Does *this accept epsilon?
// Implement Definition 3.20 and Property 3.21
int RE::Null() const {
	assert( class_invariant() );

	auto int ret;
	switch( op ) {
	case EPSILON:
        	ret = 1;
		break;
	case EMPTY:
	case SYMBOL:
        	ret = 0;
		break;
	case OR:
        	ret = left->Null() || right->Null();
		break;
	case CONCAT:
        	ret = left->Null() && right->Null();
		break;
	case STAR:
        	ret = 1;
		break;
	case PLUS:
        	ret = left->Null();
		break;
	case QUESTION:
		ret = 1;
		break;
        }
	return( ret );
}

// What is the First of *this?
// Implement Definition 4.60
CRSet RE::First() const {
	assert( class_invariant() );

	auto CRSet a;
	switch( op ) {
	case EPSILON:
	case EMPTY:
		break;
	case SYMBOL:
		a.append( sym );
		break;
	case OR:
        	a = left->First().combine( right->First() );
		break;
	case CONCAT:
		a = left->First();
		if( left->Null() ) {
			a.combine( right->First() );
		}
		break;
	case STAR:
	case PLUS:
	case QUESTION:
		a = left->First();
		break;
	}
	return( a );
}

// Is *this in similarity normal form (SNF)?
int RE::in_snf() const {
	assert( class_invariant() );

	auto int ret;
	// SNF means that all OR nodes are left-associative, and
	// in the two cases:
	//	A OR B that A < B, or
	//	(A OR B) OR C that B < C and (A OR B) is in SNF.
	switch( op ) {
	case EPSILON:
	case EMPTY:
	case SYMBOL:
		// The bases are always in SNF.
		ret = 1;
		break;
	case OR:
		if( (right->op == OR) || !right->in_snf() ) {
			// This can't be in SNF.
                        ret = 0;
		} else {
			if( left->op == OR ) {
				// Second case.
				ret = (*(left->right) < *right) && left->in_snf();
			} else {
				// First case.
                                ret = (*left < *right) && left->in_snf();
			}
                }
		break;
	case CONCAT:
        	// A CONCAT is in SNF if the subexpressions are in SNF.
        	ret = left->in_snf() && right->in_snf();
		break;
	case STAR:
	case PLUS:
	case QUESTION:
        	// A unary subexpr. is in SNF if it's subexpressions are too.
        	ret = left->in_snf();
		break;
	}
	return( ret );
}

// Put *this into similarity normal form.
RE& RE::snf() {
	assert( class_invariant() );

	switch( op ) {
	case EPSILON:
	case EMPTY:
	case SYMBOL:
        	// Already in SNF.
        	break;
	case OR:
        	// While the right is an OR, rotate them.
		// Note: this is not a true rotation, since we get
		// E | (F | G) -> (F | E) | G instead of (E | F) | G
		// but, by commutativity it's okay.
		while( right->op == OR ) {
			auto RE *temp( left );
			left = right;
			right = left->right;
                        left->right = temp;
			assert( class_invariant() );
		}
		assert( right->op != OR );
		left->snf();
		assert( class_invariant() );
		assert( left->in_snf() );
		right->snf();
		assert( class_invariant() );
		assert( right->in_snf() );
                onf();
		assert( class_invariant() );
        	break;
	case CONCAT:
		left->snf();
		assert( class_invariant() );
		assert( left->in_snf() );
		right->snf();
		assert( class_invariant() );
                assert( right->in_snf() );
        	break;
	case STAR:
	case PLUS:
	case QUESTION:
		left->snf();
		assert( class_invariant() );
		assert( left->in_snf() );
                break;
	}
        assert( in_snf() );
	assert( class_invariant() );
	return( *this );
}

// What is the derivative of *this?
// Implement Definition 5.23
RE RE::derivative( const CharRange& r ) const {
	assert( class_invariant() );

	// Duplicate some of the knowledge appearing in sig-re.cpp
	auto RE e;
	switch( op ) {
	case EPSILON:
	case EMPTY:
		assert( e.op == EMPTY );
		break;
	case SYMBOL:
		if( r <= sym ) {
			e.op = EPSILON;
			assert( e.class_invariant() );
		} else {
                	// else, it should be EMPTY.
			assert( e.op == EMPTY );
		}
                break;
	case OR:
		e.op = OR;
		e.left = new RE( left->derivative( r ) );
		e.right = new RE( right->derivative( r ) );
		assert( e.class_invariant() );
        	break;
	case CONCAT:
		if( left->Null() ) {
			e.op = OR;
			e.right = new RE( right->derivative( r ) );

			e.left = new RE;
			e.left->op = CONCAT;
			e.left->left = new RE( left->derivative( r ) );
			e.left->right = new RE( *right );
			assert( e.class_invariant() );
		} else {
			e.op = CONCAT;
			e.left = new RE( left->derivative( r ) );
			e.right = new RE( *right );
			assert( e.class_invariant() );
		}
		break;
	case STAR:
		e.op = CONCAT;
		e.left = new RE( left->derivative( r ) );
		e.right = new RE( *this );
		assert( e.class_invariant() );
                break;
	case PLUS:
		e.op = CONCAT;
		e.left = new RE( left->derivative( r ) );

		e.right = new RE;
		e.right->op = STAR;
		e.right->right = 0;
		e.right->left = new RE( *left );

		assert( e.class_invariant() );
		break;
	case QUESTION:
		e = left->derivative( r );
		break;
        // By the class_invariant(), the default: cannot occur.
	}
	assert( e.class_invariant() );
        return( (RE)e );
}

// Implement rules 1-4 of Remark 5.32 of the Taxonomy.
RE& RE::reduce() {
	assert( class_invariant() );
	switch( op ) {
	case EPSILON:
	case EMPTY:
	case SYMBOL:
		// These basis operators cannot be reduced further.
                break;
	case OR:
		left->reduce();
		assert( class_invariant() );
		right->reduce();
		assert( class_invariant() );
		if( left->op == EMPTY ) {
			// EMPTY OR E == E
			delete left;
			assimilate_node( right );
			assert( class_invariant() );
		} else if( right->op == EMPTY ) {
			delete right;
			assimilate_node( left );
			assert( class_invariant() );
		}
		// The case where left and right are both EMPTY is handled correctly.
                break;
	case CONCAT:
		left->reduce();
		assert( class_invariant() );
		right->reduce();
		assert( class_invariant() );
		if( left->op == EMPTY ) {
			// EMPTY CONCAT E == EMPTY
			delete right;
			assimilate_node( left );
			assert( class_invariant() );
		} else if( right->op == EMPTY ) {
			delete left;
			assimilate_node( right );
			assert( class_invariant() );
		} else if( left->op == EPSILON ) {
                	// EPSILON CONCAT E == E
			delete left;
			assimilate_node( right );
			assert( class_invariant() );
		} else if( right->op == EPSILON ) {
			delete right;
			assimilate_node( left );
			assert( class_invariant() );
                }
		break;
	case STAR:
		left->reduce();
		assert( class_invariant() );
		if( left->op == EPSILON ) {
                	// EPSILON * == EPSILON
			assimilate_node( left );
			assert( class_invariant() );
		} else if( left->op == EMPTY ) {
			// EMPTY * == EPSILON
			delete left;
			op = EPSILON;
			left = right = 0;
			assert( class_invariant() );
		}
                break;
	case PLUS:
		left->reduce();
		assert( class_invariant() );
		if( left->op == EPSILON ) {
			// EPSILON + == EPSILON
			assimilate_node( left );
			assert( class_invariant() );
		} else if( left->op == EMPTY ) {
			// EMPTY + == EMPTY
			assimilate_node( left );
			assert( class_invariant() );
		}
                break;
	case QUESTION:
		left->reduce();
		assert( class_invariant() );
		if( left->op == EPSILON ) {
			// EPSILON ? == EPSILON
			assimilate_node( left );
			assert( class_invariant() );
		} else if( left->op == EMPTY ) {
			// EMPTY ? == EPSILON
			delete left;
			op = EPSILON;
			left = right = 0;
			assert( class_invariant() );
		}
                break;
        }
	assert( class_invariant() );
	return( *this );
}

// Put *this into OR normal form (ONF) (used in the snf functions)
RE& RE::onf() {
	assert( class_invariant() );
	assert( op == OR );
	assert( right->op != OR );

	// Two possibilities: (A OR B) OR C, A OR B.
	if( left->op == OR ) {
		// First one.
		// Figure out which order B,C should be in:
		auto int ord( left->right->ordering( *right ) );
		if( ord == 0 ) {
			// B == C
                        // destroy C.
			delete right;
			assimilate_node( left );
		} else if( ord > 0 ) {
			// B > C
			// Need to swap B,C
			// Just do it with pointers:
			auto RE *const temp( right );
			right = left->right;
			left->right = temp;

			// Also make sure that (what's now) (A OR C)
			// is also in ONF.
                        left->onf();
		}
                // else, it's already in OR NF.
	} else {
		// Second case.
		// Figure out the order of A,B:
		auto int ord( left->ordering( *right ) );
		if( ord == 0 ) {
			// A == B
			// Wipe out one of the two nodes, and make *this
			// equal to the remaining one.
			delete right;
			assimilate_node( left );
		} else if( ord > 0 ) {
			// A > B
			// Need to swap A,B (left and right):
			auto RE *const temp( right );
			right = left;
			left = temp;
		}
                // else, it's already in OR NF
        }

	assert( class_invariant() );
	assert( in_snf() );
	return( *this );
}

// Some ordering members (largely used in similarity)
// (This is a kind of lexicographic ordering.)
int RE::ordering( const RE& r ) const {
	assert( class_invariant() );
	assert( r.class_invariant() );
	if( op != r.op ) {
		return( op - r.op );
	} else {
		assert( op == r.op );
		auto int ret;
		switch( op ) {
		case EPSILON:
		case EMPTY:
			// These two are unordered (equal).
			ret = 0;
			break;
		case SYMBOL:
			// Both are symbols, so order on the sym.
			ret = sym.ordering( r.sym );
			break;
		case OR:
		case CONCAT:
			ret = left->ordering( r.left_subexpr() );
			if( ret != 0 ) {
				// Do nothing.
			} else {
				// Compare the right subexprs, since the left
				// ones are equal.
				ret = right->ordering( r.right_subexpr() );
			}
			break;
		case STAR:
		case PLUS:
		case QUESTION:
			ret = left->ordering( r.left_subexpr() );
			break;
		}
		return( ret );
	}
}

// Assimilate *r into *this, and wipe out *r.
void RE::assimilate_node( RE *const r ) {
	// We can't assert( class_invariant() );
	//	since we can't be sure that *this is still good.
	//	This just delays catching any potential invariant violations.
        assert( r->class_invariant() );
	op = r->op;
	left = r->left;
	right = r->right;
	sym = r->sym;
	assert( class_invariant() );

	r->op = EMPTY;
	r->left = r->right = 0;
	delete r;
}

