/* (c) Copyright 1994 by Bruce Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:56:25 $
#ifndef DFA_H
#define DFA_H

#include "faabs.h"
#include "state.h"
#include "stateset.h"
#include "dtransre.h"
#include "dfacomp.h"
#include "st-eqrel.h"
#include "symrel.h"
#include "dsdfarev.h"
#include <assert.h>
#include <iostream.h>

// Deterministic finite automata: can be constructed from non-det. finite
// automata, can be minimized.

class DFA : virtual public FAabs {
public:
	// Constructors, destructors, operator=:

	// Default copy constr, destr, operator= are okay.
        DFA();

	// A special constructor used for subset construction etc.
        DFA( const DFA_components& r );

	// Some member functions:


	// A special one, to reconstruct a DFA, given new components, etc.
	inline DFA& reconstruct( const DFA_components& r );

	// Member functions required by class FAabs:

        virtual int num_states() const;
	virtual void restart();
	virtual void advance( char a );
	virtual int in_final() const;
	virtual int stuck();
	virtual DFA determinism() const;


	// A function to reverse *this.
	DFA& reverse();

	// Some minimization algorithms:

        inline DFA& min_Brzozowski();
	DFA& min_HopcroftUllman();
	DFA& min_dragon();
	DFA& min_Hopcroft();
	DFA& min_Watson();


	// Some member functions relating to useful State's.

        // Can all States reach a final State?
	int Usefulf() const;

        // Remove any States that cannot reach a final State.
	DFA& usefulf();


	// Special member functions:

	friend ostream& operator<<( ostream& os, const DFA& r );

	inline int class_invariant() const;

protected:
	// Given a minimizing equivalence relation, shrink the DFA.
	DFA& compress( const StateEqRel& r );
	DFA& compress( const SymRel& r );

	// Attempt to split the eq. class [p]_P w.r.t. [q]_P
	// (Return 1 if it was split, 0 otherwise.)
	State split( const State p, const State q, const CharRange a, StateEqRel& P ) const;

	// A helper for min_Watson.
	int are_eq( State p, State q, SymRel& S, const StateEqRel& H, const SymRel& Z ) const;

	// Implementation details:

	StatePool Q;
	// S must be a singleton set, or empty.
	StateSet S;
	StateSet F;
	DTransRel T;

	// Some simulation details:

	State current;
};

// Create a DFA with no State's, accepting nothing.
inline DFA::DFA() :
		Q(),
		S(),
		F(),
		T(),
		current( Invalid ) {
	assert( class_invariant() );
}

// A special constructor used for subset construction etc.
//	A DFA constructed this way will always have a start State.
inline DFA::DFA( const DFA_components& r ) :
		Q( r.Q ),
                S( r.S ),
		T( r.T ),
		F( r.F ) {
	current = Invalid;
	assert( class_invariant() );
}

inline DFA& DFA::reconstruct( const DFA_components& r ) {
	Q = r.Q;
	S = r.S;
	T = r.T;
	F = r.F;
	current = Invalid;
	assert( class_invariant() );
        return( *this );
}

// See Section 2 (p. 5) of the minimization Taxonomy.
inline DFA& DFA::min_Brzozowski() {
	assert( class_invariant() );
        return( reverse().reverse() );
}

inline int DFA::class_invariant() const {
	return( Q.size() == S.domain()
		&& Q.size() == F.domain()
		&& Q.size() == T.domain()
		&& current < Q.size()
		&& S.size() <= 1 );
} 


#endif
