/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:57:52 $
#ifndef DSRFA2_H
#define DSRFA2_H

#include "charrang.h"
#include "crset.h"
#include "stateset.h"
#include "staterel.h"
#include "transrel.h"
#include <assert.h>
#include <iostream.h>


// This class is used to represent abstract States in a DFA that is still under
// construction. It is used in the construction of a DFA from an RFA. The behaviour
// of a DSRFA2 follows the left-biased interpretation of RFA's (unlike DSRFA, which
// follows the right-biased interpretation). See the last item on p. 73 of the
// Taxonomy.

class DSRFA2 {
public:
	// Must always have an argument-less constructor.
	inline DSRFA2();

	inline DSRFA2( const DSRFA2& r );

	// A special constructor (for use in constructing the start state):
	DSRFA2( const Trans *rQmap_inverse,
			const StateRel *rfollow,
			const StateSet *rfirst,
			const StateSet *rlast,
			const int rfinalness );

	inline const DSRFA2& operator=( const DSRFA2& r );

	// The required member functions:
	inline int final() const;
	CRSet out_labels() const;
	DSRFA2 out_transition( const CharRange a ) const;
	inline int operator==( const DSRFA2& r ) const;
	inline int operator!=( const DSRFA2& r ) const;


	friend ostream& operator<<( ostream& os, const DSRFA2& r );

	inline int class_invariant() const;

private:
	// Another special constructor, for use in out_transitions():
	DSRFA2( const StateSet& rq,
			const Trans *rQmap_inverse,
			const StateRel *rfollow,
			const StateSet *rfirst,
                        const StateSet *rlast );

	// This stuff may vary from object to object.
	StateSet which;
	int finalness;

	// This stuff should be the same for all objects corresponding
	// to a particular RFA.
	const Trans *Qmap_inverse;
	const StateRel *follow;
	const StateSet *first;
        const StateSet *last;
};

// Must always have an argument-less constructor.
inline DSRFA2::DSRFA2() :
		which(),
		Qmap_inverse( 0 ),
		follow( 0 ),
		first( 0 ),
                last( 0 ),
		finalness( 0 ) {
	// No assert since it won't satisfy the class invariant.
}

inline DSRFA2::DSRFA2( const DSRFA2& r ) :
		which( r.which ),
		Qmap_inverse( r.Qmap_inverse ),
		follow( r.follow ),
		first( r.first ),
                last( r.last ),
		finalness( r.finalness ) {
	assert( class_invariant() );
}

inline const DSRFA2& DSRFA2::operator=( const DSRFA2& r ) {
	assert( r.class_invariant() );
	// *this may not satisfy the class invariant yet.
	which = r.which;
	Qmap_inverse = r.Qmap_inverse;
	follow = r.follow;
	first = r.first;
        last = r.last;
	finalness = r.finalness;
	assert( class_invariant() );
        return( *this );
}

inline int DSRFA2::final() const {
	assert( class_invariant() );
	return( finalness );
}

inline int DSRFA2::operator==( const DSRFA2& r ) const {
	assert( class_invariant() );
	assert( r.class_invariant() );
	assert( Qmap_inverse == r.Qmap_inverse
		&& follow == r.follow
		&& first == r.first
		&& last == r.last );
	return( (which == r.which) && (finalness == r.finalness) );
}

inline int DSRFA2::operator!=( const DSRFA2& r ) const {
	return( !operator==( r ) );
}              

inline int DSRFA2::class_invariant() const {
	return( Qmap_inverse != 0
		&& follow != 0
		&& first != 0
                && last != 0
		&& which.domain() == Qmap_inverse->range()
		&& which.domain() == follow->domain()
		&& which.domain() == first->domain()
                && which.domain() == last->domain() );
}

#endif
