/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:58:02 $
#include "fa.h"
#include "dfaseed.h"
#include "dsfa.h"
#include "dfa.h"

// Construct an FA accepting nothing.
FA::FA() : Q() {
	auto State s( Q.allocate() );
	auto State f( Q.allocate() );
	S.set_domain( Q.size() );
	S.add( s );

	F.set_domain( Q.size() );
	F.add( f );

	Transitions.set_domain( Q.size() );
	E.set_domain( Q.size() );

	current.set_domain( Q.size() );
	assert( class_invariant() );
}

FA::FA( const FA& r ) :
		Q( r.Q ),
		S( r.S ),
		F( r.F ),
		Transitions( r.Transitions ),
		E( r.E ),
		current( r.current ) {
	assert( class_invariant() );
}

FA::FA( const RE& e ) {
	// First, allocate enough states by setting up the domains.
	// See Section 4.1.1 of the Taxonomy.
	auto int states_required( states_reqd( e ) );
	S.set_domain( states_required );
	F.set_domain( states_required );
	Transitions.set_domain( states_required );
	E.set_domain( states_required );
	current.set_domain( states_required );

	auto State s( Q.allocate() );
	auto State f( Q.allocate() );
	S.add( s );
	F.add( f );
	td( s, e, f );
	assert( class_invariant() );
}

// The following follows directly from inspecting Thompson's construction
// (Definition 4.1).
int FA::states_reqd( const RE& e ) {
	assert( e.class_invariant() );

	auto int ret;
	switch( e.root_operator() ) {
	case EPSILON:
	case EMPTY:
	case SYMBOL:
		ret = 2;
		break;
	case OR:
		ret = 2 + states_reqd( e.left_subexpr() ) + states_reqd( e.right_subexpr() );
		break;
	case CONCAT:
		ret = states_reqd( e.left_subexpr() ) + states_reqd( e.right_subexpr() );
		break;
	case STAR:
	case PLUS:
        case QUESTION:
		ret = 2 + states_reqd( e.left_subexpr() );
		break;
        }
	return( ret );
}

int FA::num_states() const {
	return( Q.size() );
}

// The simulation details follow directly from Section 2 of the Taxonomy.
void FA::restart() {
	assert( class_invariant() );
	current = E.closure( S );
	assert( class_invariant() );
}

void FA::advance( char a ) {
	assert( class_invariant() );
	// Compute the epsilon-closure.
	current = E.closure( Transitions.image( current, a ) );
	assert( current == E.closure( current ) );
        assert( class_invariant() );
}

int FA::in_final() const {
	assert( class_invariant() );
	return( current.not_disjoint( F ) );
}

int FA::stuck() {
	assert( class_invariant() );
	return( current.empty() );
}

DFA FA::determinism() const {
	// Make sure that *this is structurally sound.
	assert( class_invariant() );
	// Now construct the DFA components.
	return( DFA( construct_components( DSFA( S, &Transitions, &E, &F ) ) ) );
}

ostream& operator<<( ostream& os, const FA& r ) {
	assert( r.class_invariant() );
	os << "\nFA\n";
	os << "Q = " << r.Q << '\n';
	os << "S = " << r.S << '\n';
	os << "F = " << r.F << '\n';
	os << "Transitions = \n" << r.Transitions << '\n';
	os << "E = \n" << r.E << '\n';
	os << "current = " << r.current << '\n';
	return( os ); 
}

void FA::td( const State s, const RE& e, const State f ) {
	assert( e.class_invariant() );
	// Implement function td (of Construction 4.5).
	// Construct an FA to accept the language of e, between s and f.

	switch( e.root_operator() ) {
	case EPSILON:
		E.union_cross( s, f );
                break;
	case EMPTY:
        	break;
	case SYMBOL:
		Transitions.add_transition( s, e.symbol(), f );
                break;
	case OR:
		{
			auto State p( Q.allocate() );
			auto State q( Q.allocate() );
        	        td( p, e.left_subexpr(), q );
			auto State r( Q.allocate() );
			auto State t( Q.allocate() );
			td( r, e.right_subexpr(), t );

			E.union_cross( s, p );
			E.union_cross( s, r );

			E.union_cross( q, f );
			E.union_cross( t, f );
                }
		break; 
	case CONCAT:
        	{
			auto State p( Q.allocate() );
			auto State q( Q.allocate() );
			td( s, e.left_subexpr(), p );
			td( q, e.right_subexpr(), f );

			E.union_cross( p, q );
                }
                break;
	case STAR:
        	{
			auto State p( Q.allocate() );
			auto State q( Q.allocate() );
			td( p, e.left_subexpr(), q );

			E.union_cross( s, p );
			E.union_cross( q, p );
			E.union_cross( q, f );
			E.union_cross( s, f );
                }
                break;
	case PLUS:
        	{
			auto State p( Q.allocate() );
			auto State q( Q.allocate() );
			td( p, e.left_subexpr(), q );

			E.union_cross( s, p );
			E.union_cross( q, p );
			E.union_cross( q, f );
                }
		break;
	case QUESTION:
        	{
			auto State p( Q.allocate() );
			auto State q( Q.allocate() );
			td( p, e.left_subexpr(), q );

			E.union_cross( s, p );
			E.union_cross( q, f );
			E.union_cross( s, f );
                }
		break;
        }
}

void FA::reincarnate() {
	Q.reincarnate();
	S.reincarnate();
	F.reincarnate();
	Transitions.reincarnate();
	E.reincarnate();
	current.reincarnate();
}
