/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:58:50 $
#ifndef ISIMPL_H
#define ISIMPL_H

#include "charrang.h"
#include "crset.h"
#include "re.h"
#include "bitvec.h"
#include <assert.h>
#include <iostream.h>


// This class is used to represent abstract States in a DFA that is still under
// construction. It is used as the base class in representing item sets, for the
// item set construction, DeRemer's construction, and the optimized item set
// construction. This implements Section 5.5 of the Taxonomy.

class ISImpl {
protected:
	// Must always have an argument-less constructor.
	inline ISImpl();

	// A special constructor:
	ISImpl( const RE *r );

	const ISImpl& operator=( const ISImpl& r );

	// The required member functions:
	inline int final() const;
	CRSet out_labels() const;

	inline int operator==( const ISImpl& r ) const;
	inline int operator!=( const ISImpl& r ) const;

        // Move the dots across sym nodes.
	inline void move_dots( const CharRange& r );


	friend ostream& operator<<( ostream& os, const ISImpl& r );

	inline int class_invariant() const;

	// Some implementation details:
        // Which RE is this with-respect-to:
	const RE *e;

	// Where are the dots:
	//	The indices in the BitVec's are the pre-order traversal orders of
	//	the nodes in *e.
	//	before indicates the nodes with a dot before them;
	//	after indicates the nodes with a dot after them. 
	BitVec before;
	BitVec after;

private:
	// Some helpers:

	// Move the dots as far as possible.
        //	This is named according to Definition 5.63 of the Taxonomy.
	void D( const RE& r, int& node_num );

	// Fetch the labels throughout the tree.
        // Helper to out_labels().
        CRSet traverse_labels( const RE& r, int& node_num ) const;

	// Move the dots throughout the RE tree:
        // Helper to move_dots().
	void ISImpl::traverse_move_dots( const RE& r, const CharRange& a, int& node_num );
};

// Must always have an argument-less constructor.
//	The resulting ISImpl won't satisfy the class_invariant()
inline ISImpl::ISImpl() :
		e( 0 ) {
}

inline int ISImpl::final() const {
	assert( class_invariant() );
	// It will be final if there is a dot after the root (which is node 0
	// in the traversal order).
        // See construction 5.69 of the Taxonomy.
	return( after.contains( 0 ) );
}

inline CRSet ISImpl::out_labels() const {
	assert( class_invariant() );
	// Start at the root, node 0.
	auto int node( 0 );
        return( traverse_labels( *e, node ) );
}

inline int ISImpl::operator==( const ISImpl& r ) const {
	assert( class_invariant() );
	assert( r.class_invariant() );
        // *this and r must be relative to the same RE.
        assert( e == r.e );
	// All dots must be in the same places.
	return( (before == r.before) && (after == r.after) );
}

inline int ISImpl::operator!=( const ISImpl& r ) const {
	return( !operator==( r ) );
}              

// Move the dots across sym nodes.
inline void ISImpl::move_dots( const CharRange& r ) {
	assert( class_invariant() );
	// Traverse the RE tree.
	// i maintains the node number.
	auto int i( 0 );
	// Start without dots after the nodes.
        after.clear();
	traverse_move_dots( *e, r, i );
        assert( i == e->num_operators() );

	// Now go back to the root, and do the closure.
        i = 0;
	D( *e, i );
	assert( i == e->num_operators() );
}

inline int ISImpl::class_invariant() const {
	return( e->class_invariant()
		&& before.width() == e->num_operators()
		&& after.width() == before.width() );
}

#endif
