/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:59:01 $
#include "state.h"
#include "crset.h"
#include "symrel.h"
#include "st-eqrel.h"
#include "dfa.h"
#include <assert.h>

// The following is function equiv from Section 4.6 of the min. taxonomy.
int DFA::are_eq( State p, State q, SymRel& S, const StateEqRel& H, const SymRel& Z ) const {
	if( S.contains_pair( p, q ) || H.equivalent( p, q ) ) {
		// p and q are definitely equivalent.
		return( 1 );
	} else if( !Z.contains_pair( p, q ) ) {
		assert( !S.contains_pair( p, q ) );
		assert( !H.equivalent( p, q ) );
		// p and q were already compared (in the past) and found to be
		// inequivalent.
		return( 0 );
	} else {
		assert( !H.equivalent( p, q ) );
		assert( Z.contains_pair( p, q ) );

		// Now figure out all of the valid out-transitions from p and q.
		auto CRSet C( T.out_labels( p ) );
		C.combine( T.out_labels( q ) );

		// Just make sure that p and q have the same out-transitions.
		auto int it;
		for( it = 0; !C.iter_end( it ); it++ ) {
			if( (T.transition_on_range( p, C.iterator( it ) ) == Invalid )
					|| (T.transition_on_range( q, C.iterator( it ) )
						== Invalid ) ) {
				// There's something that one transitions on, but
				// the other doesn't
				return( 0 );
			} // if
		} // for

		// p and q have out-transitions on equivalent labels.
		// Keep track of whether checking needs to continue.
		S.add_pair( p, q );
		for( it = 0; !C.iter_end( it ); it++ ) {
			auto State pdest( T.transition_on_range( p, C.iterator( it ) ) );
			auto State qdest( T.transition_on_range( q, C.iterator( it ) ) );
			if( !are_eq( pdest, qdest, S, H, Z ) ) {
				// p and q have been found distinguished.
				S.remove_pair( p, q );
				return( 0 );
			}
		} // for

		// p and q have been found equivalent.
		S.remove_pair( p, q );
		return( 1 );
	} // if
}

DFA& DFA::min_Watson() {
	assert( class_invariant() );
	assert( Usefulf() );

	// (Symmetrical) State relation S is from p.14 of the min. taxonomy.
	auto SymRel S;
	S.set_domain( Q.size() );

	// H is used to accumulate equivalence relation E.
	auto StateEqRel H( Q.size() );
	// Start with the identity since this is approximation from below
	// w.r.t. refinement.
	H.identity();

	// Z is a SymRel containing pairs of States still to be considered.
	auto SymRel Z;
	Z.set_domain( Q.size() );
	Z.identity();
	// We will need the set of non-final States to initialize Z.
	auto StateSet nonfinal( F );
	nonfinal.complement();
	Z.add_pairs( F, nonfinal );
	// Z now contains those pairs that definitely do not need comparison.

	Z.complement();
	// Z initialized properly now.

	auto State p;
	// Consider each p.
	for( p = 0; p < Q.size(); p++ ) {
		auto State q;
		// Consider each q that p still needs to be compared to.
		for( Z.image( p ).iter_start( q ); !Z.image( p ).iter_end( q );
				Z.image( p ).iter_next( q ) ) {
			// Now compare p and q.
			if( are_eq( p, q, S, H, Z ) ) {
				// p and q are equivalent.
				H.equivalize( p, q );
			} else {
				// Don't do anything since we aren't computing
				// distinguishability explicitly.
			}

			// Comparing q to p is the same as comparing [q] and [p]
			// all at once.
			// Mark them as such.
			Z.remove_pairs( H.equiv_class( p ), H.equiv_class( q ) );
		} // for
	} // for

	compress( H );

	assert( class_invariant() );
        return( *this );
}

