/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:59:05 $
#include "crset.h"
#include "state.h"
#include "stateset.h"
#include "symrel.h"
#include "dfa.h"

// Implement (a version of) Algorithm 4.7 of the minimization Taxonomy.
DFA& DFA::min_HopcroftUllman() {
	assert( class_invariant() );
	assert( Usefulf() );

	// We need the combination of all transition labels, to iterate over
	// transitions.
	auto State st;
	auto CRSet C;
	for( st = 0; st < Q.size(); st++ ) {
		C.combine( T.out_labels( st ) );
	}

	// First, figure out which are the non-final States.
	auto StateSet nonfinal( F );
	nonfinal.complement();

	// Use Z to keep track of the pairs that still need to be considered for distinguishedness.
	auto SymRel Z;
	Z.set_domain( Q.size() );

	// We begin with those pairs that are definitely distinguished.
	//	this includes pairs with different parity.
	Z.add_pairs( F, nonfinal );

	// It also includes pairs with differing out-transitions.
	//	iterate over C (the CRSet) and check the haves and have-nots
	auto StateSet have;
	have.set_domain( Q.size() );
	auto StateSet havenot;
	havenot.set_domain( Q.size() );

	auto int it;
	for( it = 0; !C.iter_end( it ); it++ ) {
		// have and havenot must be empty for the update to be correct.
		assert( have.empty() && havenot.empty() );
		auto State p;
		// Iterate over the States and check which have a transition.
		for( p = 0; p < Q.size(); p++ ) {
			// Does p have the transition?
			if( T.transition_on_range( p, C.iterator( it ) ) != Invalid ) {
				have.add( p );
			} else {
				havenot.add( p );
			}
		}
		// have and havenot are distinguished from one another.
		//	(under the assumption that Usefulf() holds)
		Z.add_pairs( have, havenot );

		have.clear();
		havenot.clear();
	}

	// G will be use to accumulate the (distinguishability) relation D.
	auto SymRel G;
	G.set_domain( Q.size() );

	// Now consider all of the pairs until nothing changes.
	while( 1 ) {
		auto State p;
		// Go looking for the next pair to do.
		for( p = 0; p < Q.size() && Z.image( p ).empty(); p++ );

		// There may be nothing left to do.
		if( p == Q.size() ) {
			break;
		} else {
			assert( !Z.image( p ).empty() );
			// Choose q such that {p,q} is in Z.
			//	We know that such a q exists.
			auto State q( Z.image( p ).smallest() );
			assert( q != Invalid );

			// Move {p,q} from Z across to G.
			Z.remove_pair( p, q );
			G.add_pair( p, q );

			// Now, check out the reverse transitions from p and q.
			auto int i;
			// Iterate over all of the labels.
			for( i = 0; !C.iter_end( i ); i++ ) {
				auto StateSet pprime( T.reverse_transition( p,
						C.iterator( i ) ) );
				auto StateSet qprime( T.reverse_transition( q,
						C.iterator( i ) ) );
				// pprime and qprime are all distinguished.
				// Iterate over both sets and flag them as distinguished.
				auto State pp;
				for( pprime.iter_start( pp ); !pprime.iter_end( pp );
						pprime.iter_next( pp ) ) {
					auto State qp;
					for( qprime.iter_start( qp );
							!qprime.iter_end( qp );
							qprime.iter_next( qp ) ) {
						// Mark pp, qp for processing if they
						// haven't already been done.
						if( !G.contains_pair( pp, qp ) ) {
							Z.add_pair( pp, qp );
						} // if
					} // for
				} // for
			} // for
		} // if
	} // while

	// Now, compute E from D
	G.complement();
	// And use it to compress the DFA.
	compress( G );

	assert( class_invariant() );
        return( *this );
}

