/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:59:19 $
#include "re.h"
#include "dsre.h"
#include "dsre-opt.h"
#include "dfaseed.h"


RE::RE( const RE& e ) :
		op( e.op ),
		right( 0 ) {	// Only the right on purpose.
	assert( e.class_invariant() );

	switch( e.root_operator() ) {
	case EPSILON:
	case EMPTY:
		left = 0;
		break;
	case SYMBOL:
		left = 0;
		// Now it's safe to copy the sym.
		sym = e.sym;
		break;
	case OR:
	case CONCAT:
               	right = new RE( e.right_subexpr() );
	case STAR:
	case PLUS:
	case QUESTION:
		left = new RE( e.left_subexpr() );
		break;
	default:
		// Should throw() something!
                break;
	}

	assert( class_invariant() );
}


RE::~RE() {
	delete left;
	delete right;
}

const RE& RE::operator=( const RE& e ) {
	assert( e.class_invariant() );

	op = e.op;
        right = 0;

	switch( e.root_operator() ) {
	case EPSILON:
	case EMPTY:
		left = 0;
		break;
	case SYMBOL:
        	left = 0;
		// In no other case is it safe to copy the sym.
		sym = e.sym;
                break;
	case OR:
	case CONCAT:
        	delete right;
        	right = new RE( e.right_subexpr() );
	case STAR:
	case PLUS:
	case QUESTION:
        	delete left;
		left = new RE( e.left_subexpr() );
		break;
	}

	assert( class_invariant() );
	return( *this );
}

// How many symbols are there in *this RE?
int RE::num_symbols() const {
	assert( class_invariant() );

	auto int ret;
	switch( op ) {
	case EPSILON:
	case EMPTY:
		ret = 0;
		break;
	case SYMBOL:
		ret = 1;
		break;
	case OR:
	case CONCAT:
		ret = left->num_symbols() + right->num_symbols();
		break;
	case STAR:
	case PLUS:
	case QUESTION:
		ret = left->num_symbols();
		break;
	}
	return( ret );
}

// How many operators are there in *this RE?
int RE::num_operators() const {
	assert( class_invariant() );

	auto int ret;
	switch( op ) {
	case EPSILON:
	case EMPTY:
	case SYMBOL:
		ret = 1;
		break;
	case OR:
	case CONCAT:
		ret = left->num_operators() + right->num_operators() + 1;
		break;
	case STAR:
	case PLUS:
	case QUESTION:
		ret = left->num_operators() + 1;
		break;
	}
	return( ret );
}

// Copy *this into *r
void RE::shallow_copy( RE *const r ) const {
	r->op = op;
	r->left = left;
	r->right = right;
	r->sym = sym;
}

void RE::reincarnate() {
	if( left != 0 ) {
		// This is a non-basic operator for reincarnation.
                // Delete what used to be there.
		delete left;
                left = 0;
		if( right != 0 ) {
                	// This is a binary operator for reincarnation.
			delete right;
                        right = 0;
		}
	}
	op = EMPTY;
	assert( class_invariant() );
}

// Output in prefix notation.
// Part of this could be REops, if enums could be used for overload resolution.
ostream& operator<<( ostream& os, const RE& r ) {
	switch( r.root_operator() ) {
	case EMPTY:
		os << '0';
		break;
	case EPSILON:
        	os << '1';
		break;
	case SYMBOL:
        	os << r.symbol();
		break;
	case OR:
		os << "| " << r.left_subexpr() << ' ' << r.right_subexpr();
		break;
	case CONCAT:
		os << ". " << r.left_subexpr() << ' ' << r.right_subexpr();
		break;
	case STAR:
		os << "* " << r.left_subexpr();
		break;
	case PLUS:
		os << "+ " << r.left_subexpr();
		break;
	case QUESTION:
		os << "? " << r.left_subexpr();
        	break;
	}
        return( os );
}

// Input, in prefix notation. Not much error checking is done!
istream& operator>>( istream& is, RE& r ) {
	// Wipe out anything that used to be in r.
	delete r.left;
	delete r.right;
	// If something goes wrong try to leave r as the EMPTY regular expression.
	r.left = r.right = 0;
	r.op = EMPTY;

	auto char c;
	// It could be that there is nothing in th input.
	if( !(is >> c) ) {
		is.clear( ios::badbit | is.rdstate() );
	} else {
		switch( c ) {
		case '1':
			r.op = EPSILON;
			break;
		case '0':
			r.op = EMPTY;
			break;
		case '[':
			// This stuff should really be as the extraction operator
			// in class CharRange.
			is >> c;
			if( c != '\'' ) {
				cerr << "First CharRange incorrect\n";
				is.clear( ios::badbit | is.rdstate() );
			} else {
				// Input first character of the range.
				is >> c;
				auto char c2;
				is >> c2;
				if( c2 != '\'' ) {
					cerr << "First CharRange not terminated properly\n";
					is.clear( ios::badbit | is.rdstate() );
				} else {
					is >> c2;
					if( c2 != '\'' ) {
						cerr << "Second CharRange incorrect\n";
						is.clear( ios::badbit | is.rdstate() );
					} else {
						// Input first char. of second range.
						is >> c2;
						r.sym = CharRange( c, c2 );
						r.op = SYMBOL;
						is >> c;
						if( c != '\'' ) {
							cerr << "Second CharRange not"
								" terminated properly\n";
							is.clear( ios::badbit | is.rdstate() );
						}
						is >> c;
						if( c != ']' ) {
							cerr << "Close block not found\n";
							is.clear( ios::badbit | is.rdstate() );
						}
					}
				}
			}
			break;
		case '\'':
			is >> c;
			r.sym = CharRange( c );
			r.op = SYMBOL;
			is >> c;
			if( c != '\'' ) {
				cerr << "Single character terminated incorrectly\n";
				is.clear( ios::badbit | is.rdstate() );
			}
			break;
		case '|':
			r.op = OR;
			r.left = new RE;
			r.right = new RE;
			is >> *r.left;
			is >> *r.right;
			break;
		case '.':
			r.op = CONCAT;
			r.left = new RE;
			r.right = new RE;
			is >> *r.left;
			is >> *r.right;
			break;
		case '*':
		case '+':
		case '?':
			if( c == '*' ) {
				r.op = STAR;
			} else if( c == '+' ) {
				r.op = PLUS;
			} else if( c == '?' ) {
				r.op = QUESTION;
			}
			r.left = new RE;
			is >> *r.left;
			break;
		default:
			cerr << "Something went wrong\n";
			is.clear( ios::badbit | is.rdstate() );
		}
		assert( r.class_invariant() );
	}
	return( is );
}

