/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:59:21 $
#ifndef RE_H
#define RE_H

#include "charrang.h"
#include "crset.h"
#include "reops.h"
#include <assert.h>
#include <iostream.h>

// Regular expressions.
// This class is not too useful for actually constructing them (for that, see
// the Sigma-algebra class in sigma.h). Class RE functions as the terms of the
// Sigma-term algebra. Some members for manipulating them are provided.

class RE {
public:
	// Some constructors, destructors, and operator=:

	// By default, create the RE denoting the empty language.
	RE();

	RE( const RE& r );

	virtual ~RE();

	const RE& operator=( const RE& r );


	// Some basic access member functions:

	// How many SYMBOL nodes in this regular expression?
        //	Mainly for use in constructing RFA's.
	int num_symbols() const;

	// How many operators in this RE?
	//	Used in ISImpl (the item set stuff).
	int num_operators() const;

	// What is the main operator of this regular expression?
	inline REops root_operator() const;

        // What is the CharRange of this RE, if it is a SYMBOL regular expression.
	inline CharRange symbol() const;

        // What is the left RE of this RE operator (if it is a higher operator).
	inline const RE& left_subexpr() const;

        // What is the right RE of this RE operator (if it a binary operator).
	inline const RE& right_subexpr() const;


	// Some derivatives (Brzozowski's) related member functions:
	// Does *this accept epsilon?
	//	This is from Definition 3.20 and Property 3.21
	int Null() const;

	// What is the First of *this?
	//	This is from Definition 4.60
	CRSet First() const;

	// Is *this in similarity normal form (SNF)?
	int in_snf() const;

	// Put *this into similarity normal form.
	RE& snf();

	// Reduce (optimize) *this by removing useless information.
	RE& reduce();

	// What is the derivative of *this (w.r.t. r)?
	RE derivative( const CharRange& r ) const;

	// Some ordering members (largely used in similarity and comparisons)
	int ordering( const RE& r ) const;

	// Some comparisons:
	inline int operator==( const RE& r ) const;
	inline int operator!=( const RE& r ) const;
	inline int operator<( const RE& r ) const;
	inline int operator>=( const RE& r ) const;
	inline int operator>( const RE& r ) const;
	inline int operator<=( const RE& r ) const;        


	// Some extras:

        friend ostream& operator<<( ostream& os, const RE& r );
	friend istream& operator>>( istream& is, RE& r );

	inline int class_invariant() const;

protected:
	// Some protected helpers, mainly for Reg<RE>:

	inline void set_root_operator( REops r );
	inline void set_symbol( const CharRange r );

	// Make a copy of *this in *r
	void shallow_copy( RE *const r ) const;

        void reincarnate();

	// Put *this into OR normal form (ONF) (used in the snf functions)
	RE& onf();

	// Destroy node *r, moving its data into *this.
	void assimilate_node( RE *const r );


	// Some implementation details.
        REops op;
	RE *left, *right;
	CharRange sym;
};

inline RE::RE() : op( EMPTY ),
		left( 0 ),
		right( 0 ) {
	assert( class_invariant() );
}

// Some inline operators:

inline REops RE::root_operator() const {
	return( op );
}

inline CharRange RE::symbol() const {
	assert( op == SYMBOL );
	return( sym );
}

// Assume that *this is a higher (non-basis) regular operator.
inline const RE& RE::left_subexpr() const {
	assert( (op != EMPTY) && (op != EPSILON) && (op != SYMBOL) );
       	return( *left );
}

// Assume that *this is a binary regular operator.
inline const RE& RE::right_subexpr() const {
	assert( (op == OR) || (op == CONCAT) );
       	return( *right );
}

inline void RE::set_root_operator( REops r ) {
	// No class assertion here, since *this is still being constructed.
	op = r;
}

inline void RE::set_symbol( const CharRange r ) {
	assert( op == SYMBOL );
	sym = r;
}

inline int RE::operator==( const RE& r ) const {
	return( ordering( r ) == 0 );
}

inline int RE::operator!=( const RE& r ) const {
	return( ordering( r ) != 0 );
}

inline int RE::operator<( const RE& r ) const {
	return( ordering( r ) < 0 );
}

inline int RE::operator>=( const RE& r ) const {
	return( ordering( r ) >= 0 );
}

inline int RE::operator>( const RE& r ) const {
	return( ordering( r ) > 0 );
}

inline int RE::operator<=( const RE& r ) const {
	return( ordering( r ) <= 0 );
}        

inline int RE::class_invariant() const {
	switch( op ) {
	case EPSILON:
	case EMPTY:
        	return( left == 0 && right == 0 );
	case SYMBOL:
        	return( left == 0 && right == 0 );
	case OR:
	case CONCAT:
		return( left != 0 && right != 0
			&& left_subexpr().class_invariant()
			&& right_subexpr().class_invariant() );
	case STAR:
	case PLUS:
	case QUESTION:
		return( left != 0 && right == 0
			&& left_subexpr().class_invariant() );
	default:
        	return( 0 );
	}
}

#endif

