/* (c) Copyright 1994 by Bruce Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:59:35 $
#ifndef SIGMA_H
#define SIGMA_H

#include "charrang.h"
#include "reops.h"
#include "re.h"

// The signature of the Sigma-algebra is defined as template Reg. It is only
// an interface template, with most of the member functions being filled in as
// specialized versions. Given class T, the class Reg<T> is a Sigma-algebra with
// T as the carrier set. Reg is the only template required, since the regular
// expression Sigma algebra only has one sort: Reg.
// The Reg of a class T is also publicly derived from T, meaning that once
// a Reg<T> is constructed (perhaps using the Sigma-algebra operators), it can
// be used as a regular T (without slicing).
// A special constructor is provided, which constructs the homomorphic image in
// algebra Reg<T> of some regular expression. (Regular expressions are the Sigma-
// term algebra.)

template<class T>
class Reg : public T {
public:
	// Some constructors. Usually pass control back to the base class.
	Reg() : T() {}

	Reg( const Reg<T>& r ) :
			// Just pass back to the base class.
			T( (const T &)r ) {}

	// Construct the homomorphic image of regular expression r, using the Sigma-
        // algebra operators (corresponding to the operators of the regular expression).
	Reg( const RE& r ) {
        	assert( r.class_invariant() );
		homomorphic_image( r );
                assert( T::class_invariant() );
	}

	// Default destructor falls through to the ~T()

	inline const Reg<T>& operator=( const Reg<T>& r ) {
		// Call through to the base-class's operator=
		T::operator=( r );
		return( *this );
        }


	// Now the Sigma-algebra signature:

	// Sigma-algebra basis
	Reg<T>& epsilon();
	Reg<T>& empty();
	Reg<T>& symbol( const CharRange r );

	// Sigma-algebra binary ops
	Reg<T>& or( const Reg<T>& r );
	Reg<T>& concat( const Reg<T>& r );

	// Sigma-algebra unary ops
	Reg<T>& star();
	Reg<T>& plus();
	Reg<T>& question();


protected:
	// Helper for constructing the homomorphic image of a regular expression.
        //	(Can also be used as a type of copy constructor for RE's.)
	inline void homomorphic_image( const RE& r );
};


// Helper for constructing the homomorphic image of a regular expression.
//	(Can also be used as a type of copy constructor for RE's.)
template<class T>
void Reg<T>::homomorphic_image( const RE& r ) {
	assert( r.class_invariant() );

       	// Construct the homomorphic image of r in *this.
	switch( r.root_operator() ) {
	case EMPTY:
		// Default constructor T() should already leave *this accepting the
		// empty language.
		//	The following call is redundant:
		empty();
		break;
	case EPSILON:
               	// Make *this accept epsilon.
               	epsilon();
               	break;
	case SYMBOL:
               	// Make *this accept the specified symbol.
               	symbol( r.symbol() );
               	break;
	case OR:
		// First make *this accept the language of the left side of the OR
		// by constructing a T for the right subexpression.
		homomorphic_image( r.left_subexpr() );
		// Make *this accept the (left or right) language.
		or( Reg<T>( r.right_subexpr() ) );
               	break;
	case CONCAT:
               	// See OR case.
		homomorphic_image( r.left_subexpr() );
		concat( Reg<T>( r.right_subexpr() ) );
               	break;
	case STAR:
               	// See OR case.
		homomorphic_image( r.left_subexpr() );
		star();
               	break;
	case PLUS:
               	// See OR case.
		homomorphic_image( r.left_subexpr() );
		plus();
		break;
	case QUESTION:
               	// See OR case.
		homomorphic_image( r.left_subexpr() );
		question();
               	break;
	}
}

#endif

