/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:59:37 $
#include "st-eqrel.h"

// Construct the total eq. relation of domain r.
StateEqRel::StateEqRel( const int r ) :
		StateTo<StateSet *>() {
	StateTo<StateSet *>::set_domain( r );

	// Now construct a StateSet representing the total relation.
	auto StateSet *const t( new StateSet );
	t->set_domain( r );
	t->complement();

	auto State p;
	for( p = 0; p < r; p++ ) {
		StateTo<StateSet *>::map( p ) = t;
	}

	// Do not delete t!!
	assert( class_invariant() );
}

StateEqRel& StateEqRel::equivalize( const State p, const State q ) {
	assert( class_invariant() );
	// Don't do anything needless:
	if( !equivalent( p, q ) ) {
		// This is dangerous if they point to the same thing.
		assert( lookup( p ) != lookup( q ) );

		// Include q's eq. class in p's
		map( p )->set_union( *lookup( q ) );

		// Get q's eq. class relating to the p's.
		auto StateSet *oldq( lookup( q ) );
		auto StateSet *newp( lookup( p ) );
		assert( oldq != newp );

		auto State i;
		for( oldq->iter_start( i ); !oldq->iter_end( i ); oldq->iter_next( i ) ) {
			map( i ) = newp;        	
		}
		assert( equivalent( p, q ) );
        	delete oldq;
		assert( lookup( p ) == lookup( q ) );
	}

	assert( class_invariant() );
        return( *this );
}

StateEqRel& StateEqRel::split( const StateSet& r ) {
	assert( class_invariant() );
	assert( domain() == r.domain() );

	// Only split it if there's actually some splitting to do.
	//	Choose first State in r for starters.
	auto State i;
	r.iter_start( i );
	// Only do the splitting if r is wholly contained in the eq. class of it's
	// smallest element, and if the splitting will actually produce a new eq.
        // class.
	if( lookup( i )->contains( r ) && *lookup( i ) != r ) {
		// Split it.
                //	Make all of not( r )'s States equivalent.
		map( i )->remove( r );
		// Make all of r's States equivalent.
                //	Build a new StateSet for the new class.
		auto StateSet *n( new StateSet( r ) );
                // i is already the first State in r.
		for( ; !r.iter_end( i ); r.iter_next( i ) ) {
                	map( i ) = n;
                }
	}

	assert( class_invariant() );
        return( *this );
}

StateEqRel& StateEqRel::identity() {
	assert( class_invariant() );

	// We'll need the representatives later.
	auto StateSet repr( representatives() );

        // Wipe out all previous stuff.
	auto State st;
	for( st = 0; st < domain(); st++ ) {
		// There may be something to wipe out here.
		// To avoid double delete [] of the same memory, only do the
                // deletion when st is the representative of it's eq. class.
		if( repr.contains( st ) ) {
		}
		// Assign the new StateSet.
		map( st ) = new StateSet;
                // It's the emptyset, and set the domain.
		lookup( st )->set_domain( domain() );
		assert( lookup( st )->empty() );
		lookup( st )->add( st );		                
	}
	assert( class_invariant() );
	return( *this );
}

StateSet StateEqRel::representatives() const {
	assert( class_invariant() );
	// Keep track of the States to consider.
	auto StateSet consider;
	consider.set_domain( domain() );
	consider.complement();

	auto StateSet ret;
	ret.set_domain( domain() );

	auto State st;
	// Iterate over all States that must be considered.
	//	starting with all of the States.
	for( consider.iter_start( st ); !consider.iter_end( st ); consider.iter_next( st ) ) {
        	ret.add( eq_class_representative( st ) );
		// No need to look at the rest of eq. class [st].
		consider.remove( equiv_class( st ) );
	}

	assert( class_invariant() );
        return( ret );
}

void StateEqRel::set_domain( int r ) {
	assert( class_invariant() );
	StateTo<StateSet *>::set_domain( r );

	assert( class_invariant() );
}

ostream& operator<<( ostream& os, const StateEqRel& r ) {
	assert( r.class_invariant() );
	os << "\nStateEqRel\n";
	// Print out all of the equivalence classes (each only once).
	auto State st;
	for( st = 0; st < r.domain(); st++ ) {
		// If i is the representative of its eq. class, then print the class.
		if( st == r.eq_class_representative( st ) ) {
			os << *r.lookup( st ) << '\n';
                }
	}
        return( os );
}

