/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:59:39 $
#ifndef STATEEQREL_H
#define STATEEQREL_H

#include "stateset.h"
#include "stateto.h"
#include <assert.h>
#include <iostream.h>


// Implement an equivalence relation on State's. These are used mainly in the DFA
// minimization algorithms. As with other State relations, the domain must be
// maintained explicitly.

class StateEqRel : protected StateTo<StateSet *> {
public:
	// Some constructors etc:
	//	Construct the total eq. relation of domain r.
	StateEqRel( const int r );

	inline StateEqRel( const StateEqRel& r );

	inline const StateEqRel& operator=( const StateEqRel& r );


	// Some members for changing the relation:

        // Make two States equivalent:
	StateEqRel& equivalize( const State p, const State q );

	// Split an equivalence class into two (assuming that r is entirely contained
	// in a class):
	StateEqRel& split( const StateSet& r );

	// Make *this the identity relation:
	StateEqRel& identity();


	// Basic access members:

	// Are States p and q equivalent?
	inline int equivalent( const State p, const State q ) const;

	// What is the equivalence class [p]?
	inline const StateSet& equiv_class( const State p ) const;

	// What is the unique representative of eq. class [p]?
	inline State eq_class_representative( const State p ) const;

	// What is the set of representatives of equivalence classes of *this?
	StateSet representatives() const;


	// Special members:

	// Domain setting stuff:
	void set_domain( int r );
	inline int domain() const;

	 
	friend ostream& operator<<( ostream& os, const StateEqRel& r );

        inline int class_invariant() const;
};

inline StateEqRel::StateEqRel( const StateEqRel& r ) :
		StateTo<StateSet *>( r ) {
	assert( class_invariant() );
}

inline const StateEqRel& StateEqRel::operator=( const StateEqRel& r ) {
	assert( class_invariant() );
	assert( r.class_invariant() );
	StateTo<StateSet *>::operator=( r );
	assert( class_invariant() );
        return( *this );
}

inline int StateEqRel::equivalent( const State p, const State q ) const {
	assert( class_invariant() );
        assert( (0 <= p) && (p < domain()) );
	assert( (0 <= q) && (q < domain()) );
	return( lookup( p ) == lookup( q ) );
}

inline const StateSet& StateEqRel::equiv_class( const State p ) const {
	assert( class_invariant() );
	assert( (0 <= p) && (p < domain()) );
        return( *lookup( p ) );
}

inline State StateEqRel::eq_class_representative( const State p ) const {
	assert( class_invariant() );
	assert( (0 <= p) && (p < domain()) );
        return( lookup( p )->smallest() );
}
 
inline int StateEqRel::domain() const {
	return( StateTo<StateSet *>::domain() );
}

inline int StateEqRel::class_invariant() const {
	auto int result( 1 );
	auto State i;
	for( i = 0; i < domain() && result; i++ ) {
		result = (domain() == lookup( i )->domain());
		result = result && lookup( i )->class_invariant();
                // i's eq. class should contain i itself.
		result = result && lookup( i )->contains( i );
		// And that's all that lookup( i ) should contain:
                //	Iterate over lookup( i ) and check that.
		auto State j;
		for( lookup( i )->iter_start( j );
				!lookup( i )->iter_end( j ) && result;
				lookup( i )->iter_next( j  ) ) {
                	result = (lookup( i ) == lookup( j )); 
                }
	}
	return( result );
}

#endif
