/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:59:41 $
#ifndef STATEPOOL_H
#define STATEPOOL_H

#include "state.h"
#include <iostream.h>

// All states in an automaton are allocated from a StatePool. StatePool's can be
// merged together to form a larger one. (Care must be taken to rename any relations
// or sets (during merging) that depend on the one StatePool.)

class StatePool {
public:
	// By default, start allocating states at 0:
	inline StatePool();

        // A copy-constructed StatePool starts at the same point.
	inline StatePool( const StatePool& r );

	// Default destructor and operator= are okay.

        // How many states are already allocated (one more than that last allocated one,
	// since it begins at 0).
	inline int size() const;

	// Incorporate another StatePool by making it disjoint. Anything
	// associated with the other pool must be renamed explicitly.
	inline StatePool& incorporate( const StatePool r );

        // Allocate a new state.
	inline State allocate();

	// Clear all states that were allocated; start again at 0.
	inline StatePool& reincarnate();

	// Does this StatePool contain State r.
	inline int contains( const State r ) const;

        // Output a StatePool.
	friend ostream& operator<<( ostream& os, const StatePool r );

private:
	// The next one to be allocated.
	int next;
};

inline StatePool::StatePool() :
		next( 0 ) {}

inline StatePool::StatePool( const StatePool& r ) :
		next( r.next ) {}

inline int StatePool::size() const {
	return( next );
}

inline StatePool& StatePool::incorporate( const StatePool r ) {
	next += r.size();
	return( *this );
}

inline State StatePool::allocate() {
	return( next++ );
}

inline StatePool& StatePool::reincarnate() {
	next = 0;
	return( *this );
}

inline int StatePool::contains( const State r ) const {
	return( 0 <= r && r < next );
}

inline ostream& operator<<( ostream& os, const StatePool r ) {
	return( os << "[0," << r.next << ')' );
}

#endif

