/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:59:47 $
#ifndef STATEREL_H
#define STATEREL_H

#include "state.h"
#include "stateset.h"
#include "stateto.h"
#include <assert.h>
#include <iostream.h>

// Implement binary relations on States. This is most often used for epsilon transitions
// and follow relations.

class StateRel : protected StateTo<StateSet> {
public:
	// Constructors, destructors, operator=:

	// Default argument-less constructor calls to the base class.
	inline StateRel();

        // Copy constructor does too.
	inline StateRel( const StateRel& r );

        // Default destructor is okay

	inline const StateRel& operator=( const StateRel& r );


	// Some relational operators:

	// Compute the image of r under *this.
	StateSet image( const StateSet& r ) const;

	// Or, compute the image of a single State.
	inline const StateSet& image( const State r ) const;


        // Compute the reflexive and transitive closure of r under *this.
	StateSet closure( const StateSet& r ) const;

	// Some functions updating *this:

	// Member functions union_cross(A,B) makes *this the union (relation-
	// wise) of *this with A times B (Cartesian cross product).
	inline StateRel& union_cross( State p, State q );
	inline StateRel& union_cross( State st, const StateSet& S );
	StateRel& union_cross( const StateSet& S, State st );
	StateRel& union_cross( const StateSet& A, const StateSet& B );

	// Remove a pair of States from the relation.
	inline StateRel& remove_pair( const State p, const State q );
	StateRel& remove_pairs( const StateSet& P, const StateSet& Q );

	// Clear out this relation, without changing the domain.
	void clear();

	// Perform normal union of two relations.
	StateRel& set_union( const StateRel& r );


	// Some domain members:

        // What is the domain of this relation.
	inline int domain() const;

        // Change the domain of this relation.
	void set_domain( const int r );

	// Recycle this entire relation.
	void reincarnate();


        // Union relation r into *this, while adjusting r.
	StateRel& disjointing_union( const StateRel& r );


	// Some special members:

	friend ostream& operator<<( ostream& os, const StateRel& r );

	int class_invariant() const;
};

inline StateRel::StateRel() :
		StateTo<StateSet>() {
	assert( class_invariant() );
}

inline StateRel::StateRel( const StateRel& r ) :
		StateTo<StateSet>( (StateTo<StateSet> &)r ) {
	assert( class_invariant() );
}

inline const StateRel& StateRel::operator=( const StateRel& r ) {
	assert( r.class_invariant() );
	// call back to the base class.
        StateTo<StateSet>::operator=( (const StateTo<StateSet>&) r );
	assert( class_invariant() );
	return( *this );
}

inline const StateSet& StateRel::image( const State r ) const {
	assert( class_invariant() );
        assert( 0 <= r && r < domain() );
	// call back to the base class.
        return( lookup( r ) );
}

inline int StateRel::domain() const {
	return( StateTo<StateSet>::domain() );
}

inline StateRel& StateRel::union_cross( State p, State q ) {
	assert( class_invariant() );
	assert( (0 <= p) && (p < domain()) );
	assert( (0 <= q) && (q < domain()) );
	map( p ).add( q );
	assert( class_invariant() );
	return( *this );
}

// Map a st also to StateSet S.
inline StateRel& StateRel::union_cross( State st, const StateSet& S ) {
	assert( class_invariant() );
        assert( S.class_invariant() );
        assert( (0 <= st) && (st < domain()) && (S.domain() == domain()) );
	map( st ).set_union( S );
        assert( class_invariant() );
	return( *this );
}

inline StateRel& StateRel::remove_pair( const State p, const State q ) {
	assert( class_invariant() );
	assert( (0 <= p) && (p < domain()) );
	assert( (0 <= q) && (q < domain()) );
	map( p ).remove( q );
	map( q ).remove( p );
	assert( class_invariant() );
	return( *this );
}

inline int StateRel::class_invariant() const {
	// First, we must satisfy the base class_invariant().
	auto int ret( StateTo<StateSet>::class_invariant() );
	auto int i;
	for( i = 0; (i < domain()) && ret; i++ ) {
		ret = ret && lookup( i ).class_invariant()
			&& (lookup( i ).domain() == domain());
        }
	return( ret );
}

#endif

