/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:59:52 $
#ifndef STATESET_H
#define STATESET_H

#include "bitvec.h"
#include "state.h"
#include <iostream.h>
#include <assert.h>

// Implement a set of State's. Normal set operations are available. The StateSet
// is normally associated (implicitly) with a particular StatePool; whenever a StateSet
// is to interact with another (from a different StatePool), its States must be renamed
// (to avoid a name clash). The capacity of a StateSet must be explicitly managed; many
// set operations are not bounds-checked when assert() is turned off.

class StateSet : protected BitVec {
public:
	// Constructors, destructors, operator=:

	// Default constructor creates the emptyset, with little effort.
	// This assumes that delete [] 0 is okay.
	inline StateSet();

	// Normal copy constructor.
	inline StateSet( const StateSet& r );

	// Normal non-virtual destructor, default will call the base.

        // Needs a special operator=.
	inline const StateSet& operator=( const StateSet& r );


	// Tests:

	// A special equality operator.
	inline int operator==( const StateSet& r ) const;

        // Inequality operator.
	inline int operator!=( const StateSet& r ) const;

	// Is this set empty?
	inline int empty() const;

	// What is the size of this set (cardinality)?
	inline int size() const;

	// Set operators (may affect *this):

	// complement *this.
	inline StateSet& complement();

	// inserts a State.
	inline StateSet& add( const State r );

        // remove a State from the set.
	inline StateSet& remove( const State r );

	// Set union.
	inline StateSet& set_union( const StateSet& r );

        // Set intersection.
	inline StateSet& intersection( const StateSet& r );

        // Set difference.
	inline StateSet& remove( const StateSet& r );

        // Set containment.
	inline int contains( const StateSet& r ) const;

	// Is a State in the set?
	inline int contains( const State r ) const;

        // Does this set have something in common with r?
	inline int not_disjoint( const StateSet& r ) const;

	// Make this set the emptyset.
	inline void clear();

	// What is the smallest element of *this?
	inline State smallest() const;


	// Some domain related members:

	// How many States can this set contain?
        //	[0,domain()) can be contained in *this.
	inline int domain() const;

	inline void set_domain( const int r );

	// Recycle this StateSet.
	inline void reincarnate();

	// Rename the elements of this StateSet so that they don't fall within
        // StatePool r.
	inline StateSet& st_rename( const int r );

	// Include another StateSet into this one, renaming all the States.
	inline StateSet& disjointing_union( const StateSet& r );


	// Iterators:

	// Place the first State in the iteration in reference r.
        // r == Invalid if there is no first one.
	inline State iter_start( State& r ) const;

        // Is r the last State in an iteration sequence.
	inline int iter_end( State r ) const;

        // Place the next State, after r (in the iteration sequence), in reference r.
	inline State iter_next( State& r ) const;


	// Other special members:

	friend ostream& operator<<( ostream& os, const StateSet& r );

	// Structural invariant on the StateSet, just the class invariant of the base.
	inline int class_invariant() const;
};

inline StateSet::StateSet() :
		BitVec() {}

inline StateSet::StateSet( const StateSet& r ) :
		BitVec( r ) {}

inline const StateSet& StateSet::operator=( const StateSet& r ) {
	BitVec::operator=( r );
        return( *this );
}

inline int StateSet::operator==( const StateSet& r ) const {
	return( BitVec::operator==( r ) );
}

inline int StateSet::operator!=( const StateSet& r ) const {
	return( !operator==( r ) );
}

inline int StateSet::empty() const {
	return( !BitVec::something_set() );
}

inline int StateSet::size() const {
	return( BitVec::bits_set() );
}

inline StateSet& StateSet::complement() {
	assert( class_invariant() );
	BitVec::bitwise_complement();
	assert( class_invariant() );
	return( *this );
}

inline StateSet& StateSet::add( const State r ) {
	BitVec::set_bit( r );
        return( *this );
}

inline StateSet& StateSet::remove( const State r ) {
	BitVec::unset_bit( r );
        return( *this );
}

inline StateSet& StateSet::set_union( const StateSet& r ) {
	BitVec::bitwise_or( r );
        return( *this );
}

inline StateSet& StateSet::intersection( const StateSet& r ) {
	BitVec::bitwise_and( r );
        return( *this );
}

inline StateSet& StateSet::remove( const StateSet& r ) {
	BitVec::bitwise_unset( r );
        return( *this );
}

inline int StateSet::contains( const StateSet& r ) const {
	return( BitVec::contains( r ) );
}

inline int StateSet::contains( const State r ) const {
	return( BitVec::contains( r ) );
}

inline int StateSet::not_disjoint( const StateSet& r ) const {
	return( BitVec::something_common( r ) );
}

inline void StateSet::clear() {
	BitVec::clear();
}

inline State StateSet::smallest() const {
	return( BitVec::smallest() );
}

inline int StateSet::domain() const {
	return( BitVec::width() );
}

inline void StateSet::set_domain( const int r ) {
	assert( r >= BitVec::width() );
	BitVec::set_width( r );
}

inline void StateSet::reincarnate() {
	BitVec::reincarnate();
}

inline StateSet& StateSet::st_rename( const int r ) {
	BitVec::left_shift( r );
        return( *this );
}

inline StateSet& StateSet::disjointing_union( const StateSet& r ) {
	BitVec::append( r );
        return( *this );
}

inline State StateSet::iter_start( State& r ) const {
	return( BitVec::iter_start( r ) );
}

inline int StateSet::iter_end( State r ) const {
	return( BitVec::iter_end( r ) );
}

inline State StateSet::iter_next( State& r ) const {
	return( BitVec::iter_next( r ) );
}

inline int StateSet::class_invariant() const {
	// The Invalid stuff is a required definition.
	return( BitVec::class_invariant() && (Invalid == -1) );
}

#endif

