/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 15:59:54 $
#ifndef STATETO_H
#define STATETO_H

#include "state.h"
#include <assert.h>
#include <iostream.h>

// Map State's to class T's. Used to implement transitions (TransRel and DTransRel),
// and State relations (StateRel and StateEqRel).

template<class T>
class StateTo {
// Cannot be protected because the function in dfaseed.h (construct_components)
// needs access.
public:
	// Constructors, destructors, operator=:

	// Default is to not map anything.
	StateTo();

	// Copying can be costly. Use-counts could make this cheaper.
	StateTo( const StateTo<T>& r );

	// Assume delete [] 0 is okay.
	virtual ~StateTo();

	// Use counts could make this cheaper.
	const StateTo<T>& operator=( const StateTo<T>& r );

	// First, a const lookup operator.
	inline const T& lookup( const State r ) const;

	// Used to associate a State and a T in the mapping. Note: not const.
	inline T& map( const State r );


	// Some domain members:

        // How many States can *this map?
	inline int domain() const;

	// Set a new domain.
	void set_domain( const int r );

        // Note the reincarnate() doesn't reincarnate the T's.
	inline void reincarnate();


	// Allow two mappings to be combined; the domain of *this remains the same, while the
	// domain of r is renamed to not clash with the domain of *this.
	StateTo<T>& disjointing_union( const StateTo<T>& r );


	// Some extras:

	friend ostream& operator<<( ostream& os, const StateTo<T>& r );

	// Assert that everything's okay.
	inline int class_invariant() const;


private:
	// Represent the map as a dynamically allocated array of T's.
	int howmany;
        int in_use;
 	T *data;

	// When the array is grown by a certain amount, it also grows by an extra
	// buffer amount for efficiency.
	enum { expansion_extra = 5 };
};

template<class T>
StateTo<T>::StateTo() :
		howmany( 0 ),
		in_use( 0 ),
		data( 0 ) {
	assert( class_invariant() );
}

// Copying can be costly. Use-counts could make this cheaper.
template<class T>
StateTo<T>::StateTo( const StateTo<T>& r ) :
		howmany( r.in_use + expansion_extra ),
		in_use( r.in_use ),
		data( new T [r.in_use + expansion_extra] ) {
       	assert( r.class_invariant() );
	auto int i;
	for( i = 0; i < r.in_use; i++ ) {
		data[i] = r.data[i];
	}
	assert( class_invariant() );
}

// Assume there is nothing wrong with delete [] 0.
template<class T>
StateTo<T>::~StateTo() {
	delete [] data;
}

// Use-counts could make this cheaper.
template<class T>
const StateTo<T>& StateTo<T>::operator=( const StateTo<T>& r ) {
       	assert( r.class_invariant() );
	// Prevent assignment to self.
	if( this != &r ) {
               	// Is there enough memory to contain the assignment?
               	if( howmany < r.in_use ) { 
			delete [] data;
			data = new T [howmany = r.in_use + expansion_extra];
		}
		in_use = r.in_use;
		assert( in_use <= howmany );
		auto int i;
		for( i = 0; i < in_use; i++ ) {
			data[i] = r.data[i];
		}
	}
	assert( class_invariant() );
	return( *this );
}

// The actual mapping functions:
template<class T>
inline const T& StateTo<T>::lookup( const State r ) const {
       	assert( class_invariant() );
	// First check that it's in bounds.
	assert( 0 <= r && r < domain() );
	return( data[r] );
}

// Used to associate a State and a T in the mapping. Note: not const.
template<class T>
inline T& StateTo<T>::map( const State r ) {
       	assert( class_invariant() );
	// First check that it's in bounds.
	assert( 0 <= r && r < domain() );
	return( data[r] );
}

// How many States can *this map?
template<class T>
inline int StateTo<T>::domain() const {
       	return( in_use );
}

// Set a new domain.
template<class T>
void StateTo<T>::set_domain( const int r ) {
	assert( class_invariant() );
	// Make sure that things aren't shrinking.
	assert( r >= in_use );
	if( r > howmany ) {
		auto int i;
		auto T *d( new T [howmany = r + expansion_extra] );
		for( i = 0; i < in_use; i++ ) {
			d[i] = data[i];
		}
		delete [] data;
		data = d;
		// Note: the newly created T's are only default constructed.
	}
	in_use = r;
	assert( class_invariant() );
}

// NB: the reincarnate() doesn't reincarnate the T's.
template<class T>
inline void StateTo<T>::reincarnate() {
	in_use = 0;
}

// Allow two mappings to be combined; the domain of *this remains the same, while the
// domain of r is renamed to avoid this->domain().
template<class T>
StateTo<T>& StateTo<T>::disjointing_union( const StateTo<T>& r ) {
       	assert( class_invariant() && r.class_invariant() );
	// Don't do anything that's needless.
	if( r.domain() != 0 ) {
		auto int olddom( domain() );
		set_domain( domain() + r.domain() );
		assert( domain() == olddom + r.domain() );

		// Copy r into the new location.
		auto int i;
		for( i = 0; i < r.domain(); i++ ) {
                       	data[i + olddom] = r.data[i];
		}
	}
	assert( class_invariant() );
	return( *this );
}

// T is assumed to have an operator<<
template<class T>
inline ostream& operator<<( ostream& os, const StateTo<T>& r ) {
	assert( r.class_invariant() );
	auto State i;
	for( i = 0; i < r.in_use; i++ ) {
		os << i << "->" << r.data[i] << endl;
	}
	return( os );
}

// Assert that everything's okay.
template<class T>
inline int StateTo<T>::class_invariant() const {
	return( 0 <= in_use
		&& in_use <= howmany
		// So 0 <= howmany by transitivity.
		&& (howmany != 0 ? data != 0 : data == 0) );
}


#endif

