/* (c) Copyright 1994 by Bruce W. Watson */
// $Revision: 1.1 $
// $Date: 1994/05/02 16:00:44 $
#ifndef TRANS_H
#define TRANS_H

#include "charrang.h"
#include "crset.h"
#include "state.h"
#include "stateset.h"
#include "tr-pair.h"
#include "tr-impl.h"
#include <assert.h>
#include <iostreams.h>

// Map char to StateSet, using TransImpl.
// Implement the transitions for one state, or for the inverse of Qmap
// (in RFA's, LBFA's, and RBFA's).

class Trans : protected TransImpl {
public:
	// Constructors, destructors, operator=:

	// Assume that delete [] 0 is okay.
	inline Trans();

	// Copy constructor allocates more memory. Should use (use-counting for
	// efficiency.
	inline Trans( const Trans& r );


        // operator=() must copy the memory.
	inline const Trans& operator=( const Trans& r );


	// Some member functions for making transitions:

	// Map a char to the corresponding StateSet.
	StateSet operator[]( const char a ) const;

	// Map a CharRange to the corresponding StateSet
	//	assuming that the CharRange is entirely contained in the label
	//	of a transition.
        StateSet range_transition( const CharRange a ) const;

	// What are all of the transitions in *this?
	inline CRSet out_labels() const;

	// What are all transition labels into StateSet r?
	inline CRSet labels_into( const StateSet& r ) const;


	// Some special member functions:

	// Clear out all prev. transitions, and zero the domain.
	inline void reincarnate();

        // The range of States that can be transitioned to.
	inline int range() const;

	// Change the range of States that can be transitioned to.
	//	This is used in determining the domain() of the StateSet's (which
        //	are destinations of transitions).
	inline void set_range( const int r );


	// Add one more transition to the set.
	inline Trans& add_transition( const CharRange a, const State q );

	// Do normal set union.
	Trans& set_union( const Trans& r );

	// Incorporate another Trans, while renaming all of the states.
	Trans& disjointing_union( const Trans& r );
	 
	// Rename all of the States (that are destinations of some transition)
	// such that none of them fall in the range [0,r).
	Trans& st_rename( const int r );

        // Output the transitions.
        friend ostream& operator<<( ostream& os, const Trans& r );

	// Maintain the class invariant.
	inline int class_invariant() const;


private:
	// Implementation details:
	// The domain() of the associated StatePool.
	int destination_range;
};


// Some inline members:
inline Trans::Trans() :
		TransImpl(),
		destination_range( 0 ) {
	assert( class_invariant() );
}

inline Trans::Trans( const Trans& r ) :
		TransImpl( (TransImpl&)r ),
		destination_range( r.destination_range ) {
	assert( class_invariant() );
}

inline const Trans& Trans::operator=( const Trans& r ) {
	assert( class_invariant() );
	assert( r.class_invariant() );

	TransImpl::operator=( (TransImpl&)r );
        destination_range = r.destination_range;
	assert( class_invariant() );
	return( *this );
}

inline CRSet Trans::out_labels() const {
	assert( class_invariant() );
	return( TransImpl::out_labels() );
}

inline CRSet Trans::labels_into( const StateSet& r ) const {
	assert( class_invariant() );
        return( TransImpl::labels_into( r ) );
}

inline void Trans::reincarnate() {
	assert( class_invariant() );
	TransImpl::reincarnate();
	assert( class_invariant() );
}

inline int Trans::range() const {
	assert( class_invariant() );
	return( destination_range );
}

inline void Trans::set_range( const int r ) {
	assert( class_invariant() );
	destination_range = r;
	assert( class_invariant() );
}

inline Trans& Trans::add_transition( const CharRange a, const State q ) {
	assert( class_invariant() );
	TransImpl::add_transition( a, q );
	assert( class_invariant() );
	return( *this );
}

// Maintain the class invariant.
inline int Trans::class_invariant() const {
	auto int ret( 1 );
	auto int i;
	for( i = 0; (i < in_use) && ret; i++ ) {
		ret = (0 <= transitions( i ).transition_destination)
			&& (transitions( i ).transition_destination < destination_range);
	}
	return( ret );
}

#endif

